/****************************************************************/
/*              Butterfly HyperQuickSort Program                */
/*                                                              */
/*                    adapted from                              */
/*                                                              */
/* http://www.cs.rit.edu/usr/local/pub/ncs/parallel/mpi/hqs.c   */
/****************************************************************/

#include "../eSkel.c"
#include <math.h>
/* #include <sys/types.h> */

#define         HOST            0
#define         MAX_ARRAY       500000
#define         MIN_DIM         0
#define         MAX_DIM         3
#define         MAX_NODES       1 << MAX_DIM
#define         SEED            17

void merge();
void QuickSort();

int *half;

int 
        /* half[MAX_ARRAY],        Subarray for merging                   */


        lower[MAX_ARRAY],       /* Array for subdividing problem          */
        items[MAX_ARRAY],       /* Array of info to be sorted             */
        localMedian,            /* Median of all sorted substrings        */
        globalMedian,           /* Median of all sorted substrings        */
        medians[MAX_NODES],     /* Array of medians                       */
        upper[MAX_ARRAY];       /* Array for subdividing problem          */

double  startTime, endTime;     /* variables to handle MPI timer results  */

int     dimension,              /* Dimension of hypercube                 */
        halflen,                /* Length of subarray for merging         */
        i,                      /* Loop control variable                  */
        j,                      /* Loop control variable                  */
        k,                      /* Loop control variable                  */
        length,                 /* Length of this node's list             */
        lowerlen,               /* Length of subdivision                  */
        num_to_sort,            /* Size of array to sort                  */
        num_nodes,              /* Number of nodes in hypercube           */
        procnum,                /* Loop control variable                  */
        upperlen;               /* Length of subdivision                  */

MPI_Status
        Status;                 /* Status returned from MPI_Recv function */

/*******************************************************/
/* --- HQSInput --- array size and hypercube dimension */
/*******************************************************/
int HQSInput(int *num_to_sort, int *dimension)
{

   printf("\nEnter size of array: "); 
   fflush(stdout);
   scanf("%d", num_to_sort); 
   printf("\n");
   fflush(stdout);
   while ( (*num_to_sort < 2) || (*num_to_sort > MAX_ARRAY) )
   {
      printf("\nSize must be between 2 and %d\n", MAX_ARRAY);
      fflush(stdout);
      printf("\nEnter size of array: ");
      fflush(stdout);
      scanf("%d", num_to_sort);
      printf("\n");
      fflush(stdout);
   }

   printf("\nEnter dimension of cube: "); 
   fflush(stdout);
   scanf("%d", dimension); 
   while ( (*dimension < MIN_DIM) || (*dimension > MAX_DIM) )
   {
      printf("\nDimension must be between %d and %d\n", MIN_DIM, MAX_DIM);
      fflush(stdout);
      printf("\nEnter dimension of cube: ");
      fflush(stdout);
      scanf("%d", dimension); 
   }

   return 1;
}


/*********************************************************************/
/* --- SendData--- send subarray and length to each node */
/*********************************************************************/
void SendData(int num_to_sort, int num_nodes, int *length0, int items[])
{
int     
        i,                      /* Loop control variable                  */
        extras,                 /* Number of processors that get one more */
        length,                 /* Lengths of subarrays for each node     */
        length1;                /* Lengths of subarrays for each node     */


   /* Distribute array data to nodes */

   /* Initialize array to sort */
   srand(SEED);
   for (i = 0; i < num_to_sort; i++)
   {
      items[i] = rand();
   }
   printf("Data generation complete.  Sorting begins.\n\n");
   fflush(stdout);


   /* Load balance, at least initially */
   length = num_to_sort / num_nodes;
   extras = num_to_sort - length * num_nodes;
   length1 = length + 1;


   /* Send initial data to be sorted to each process */
   for (i = 1; i < num_nodes; i++)
   { 
      if (i >= extras)
      {
         MPI_Send(&items[length*i+extras], length, MPI_INT, i, 0, 
                  mycomm());
      }
      else
      {
         MPI_Send(&items[length*i+(i<extras)*i], length1, MPI_INT, i, 0, 
                  mycomm());
      }
   }     

   /* Length for processor 0 */
   if (extras)
      *length0 = length1;
   else
      *length0 = length;
}


void getdata()
{
   /* Process HOST reads input data: number of numbers to sort, hypercube
      dimension */
   if (myrank() == HOST) {
      HQSInput(&num_to_sort, &dimension);
   }

   /* Broadcast dimension, the hypercube dimension to the world */
   MPI_Bcast(&dimension, 1, MPI_INT, HOST, mycomm());

   /* Initialize variables */
   num_nodes = 1 << dimension;

    /* Verify that there are enough processor running */
    if (mycommsize() != num_nodes) 
    {
        if (myrank() == HOST){
        printf("There are %d processors executing, %d needed, quitting\n",
               mycommsize(), num_nodes);
        fflush(stdout);
        }
        MPI_Finalize();
        exit();
    }

   /* Generate numbers to sort and split the work among the processors */
   if (myrank() == HOST) 
   {
      /* SendData will return length and array for the HOST node */
      SendData(num_to_sort, num_nodes, &length, items);
   }
   else
   {
      MPI_Recv(items, MAX_ARRAY, MPI_INT, HOST, 0, mycomm(), 
          &Status);
      MPI_Get_count(&Status, MPI_INT, &length);
   }

}

void checkandreport(void)
{
  int p; 

   /* Confirm array is sorted */
   for (i = 0; i < length-1; i++)
   {
      if (items[i] > items[i+1])
      {
         printf("%d error --(%d:%d) (%d:%d)\n", myrank(), i, items[i], 
        i+1, items[i+1]);
         fflush(stdout);
      }
   }

   /* Report results */
   for (p=0; p<mycommsize(); p++) {

     if (myrank() == p) {
       if (length != 0)
       {
          printf("Node %d (time=%6.6f seconds, length = %d)  : %d ... %d\n\n",
                  myrank(), endTime, length, items[0], items[length-1]);
          fflush(stdout);
       }
       else
       {
          printf("Node %d (time=%6.6f seconds, length = %d)\n\n", myrank(), endTime, length);
          fflush(stdout);
       }
     }
     MPI_Barrier(mycomm());
   }
}


/*******************************************************************/
/* --- merge --- merge and sort two sorted arrays into third array */
/*******************************************************************/
void merge(int a[], int b[], int c[], int m, int n)
{
   int  i,      /* Loop control for array a */
        j,      /* Loop control for array b */
        k;      /* Loop control for array c */

   i = 0; j = 0; k = 0;

   while ((i < m) && (j < n))
      {
      if (a[i] < b[j])
         {
         c[k] = a[i];
         k++; i++;
         }
      else
         {
         c[k] = b[j];
         k++; j++;
         };
      }


   /* Pick up any remainder */
   while (i < m)
      {
      c[k] = a[i];
      k++; i++;
      };
   while (j < n)
      {
      c[k] = b[j];
      k++; j++;
      };

}


/********************************************************/
/*               Swap: swap two values                  */
/********************************************************/
void Swap(int *i, int *j)
{
int 
        t;              /* Temporary variable for swap */

   t = *i;
   *i = *j;
   *j = t;
}


/*****************************************************************************/
/* Median3: Find the median of left, center, and right; return the median
   as the pivot, and "hide" the pivot as the next-to-last (right-1) element. */
/*****************************************************************************/
void Median3 (int a[], int left, int right, int *pivot)
{
int 
        center;   /* Temporary variable for determining median */

   center = (left+right) / 2;

   if ( a[left] > a[center])
   {
       Swap (&a[left], &a[center]);
   }
   if ( a[left] > a[right])
   {
       Swap (&a[left], &a[right]);
   }
   if ( a[center] > a[right])
   {
       Swap (&a[center], &a[right]);
   }

   *pivot = a[center];
   Swap (&a[center], &a[right-1]);
}


/********************************************************/
/*      QSort: the basic QuickSort algorithm.           */
/********************************************************/
void QSort(int a[], int left, int right)
{
int 
        i,              /* Loop control variable         */
        j,              /* Loop control variable         */
        pivot;          /* Quicksort pivot              */

   if (left < right)
   {
      Median3 (a, left, right, &pivot); /* Find the pivot. */

      /* Do the partition. */
         i = left;
         j = right-1;
         if (i < j)
         {
           do
           {
              do
              {
                 i++;
              }
              while (a[i] < pivot);
              do
              {
                 j-- ;
              }
              while (a[j] > pivot);
              Swap(&a[i], &a[j]);
           }
           while (j > i);

           Swap(&a[i], &a[j]);  /* Undo last swap. */
           Swap(&a[i], &a[right-1]);  /* Restore pivot. */
         };

      QSort(a, left, i-1);
      QSort(a, i+1, right);
   }
}


/*******************************************************/
/*      QuickSort: standard quicksort sorting routine  */
/*******************************************************/
void QuickSort(int d[], int n)
{
   QSort(d, 0, n-1);
}


void level (void)
{

      /* calculate my median */
      if (length % 2)                   /* if odd */
      {
         localMedian = items[length / 2];
      }
      else
      {
         localMedian = (items[(length / 2) - 1] + items[length / 2]) / 2;
      }

      /* Broadcast median to each node in subcube (except me) */
      medians[myrank()] = localMedian;
      for (j = 0; j < mycommsize(); j++) {
          MPI_Bcast(&medians[j], 1, MPI_INT, j, mycomm());
      }

      /* Determine median of medians */
      QuickSort(medians, mycommsize());
      /* There are always an even number of medians */
      globalMedian = (medians[(mycommsize() / 2) - 1] + medians[mycommsize() / 2]) / 2;


      /* 4) Create lower and upper 'halves' of items */
      j = 0; k = 0;
      while ((items[j] <= globalMedian) && (j < length)) 
      {
         lower[k] = items[j];
         k++; 
         j++;
      };
      lowerlen = k; 
      k = 0;
      while (j < length) 
      {
         upper[k] = items[j];
         k++; 
         j++;
      };
      upperlen = k;


      /* Exchange info with partner */
      if (myrank() >= mycommsize()/2){
        /* Send lower 'half' to partner */
	Give (lower, lowerlen);
        halflen = MAX_ARRAY;
        half = Take (&halflen);
      } else {
        /* Receive lower 'half' from partner */
        halflen = MAX_ARRAY;
        half = Take (&halflen);
        /* Send upper 'half' to partner */
	Give (upper, upperlen);
      }


      /* 6) Merge lower/upper array with 'half' array into items array */
      if (myrank() >= mycommsize()/2) 
      {
         merge(upper, half, items, upperlen, halflen);
         length = upperlen + halflen;
      }
      else
      {
         merge(lower, half, items, lowerlen, halflen);
         length = lowerlen + halflen;
      }
      free(half);
}




/*******************************************************/
/*      Main program - starts up on each machine       */
/*******************************************************/
int main(int argc,char *argv[])
{

   /*  Set up mpi machine */
   MPI_Init(&argc, &argv);
   SkelLibInit();
    
   getdata();

   MPI_Barrier(mycomm());
   startTime = MPI_Wtime();

   /* local sort using quicksort */
   QuickSort(items, length);


   Butterfly (dimension, level, MPI_INT, mycomm());

   MPI_Barrier(MPI_COMM_WORLD);
   endTime = MPI_Wtime() - startTime;

   checkandreport();

   MPI_Finalize();
   return 0;
}
