void Pipeline1for1v (int ns, eSkel_atom_t *(*stages[])(eSkel_atom_t *),
                    int col, spread_t spr[], MPI_Datatype ty[],
                    void *in, int *inlen, int inmul,
                    void *out, int *outlen, int *outmul, MPI_Comm comm)
{
  /* Variant in which Give and Take are implicit, stage is a one-for-one function   */

  /* outlens[0] is the length, in units big enough for ty[ns], of the out buffer,    */
  /* ie what is the maximum accumulated output length we can handle                  */
  /* outlens[1] is the length, in units big enough for an int, of the outlens buffer */
  /* ie what is the maximum outmult we can handle                                    */

  int i, j, k, l, m, n, temp, 
      outbuffroot,        /* the root process in final stage as an id in comm */
      *tempblens;         /* complete item lengths on st 0 p0 */
  char **tempbuff;        /* temporary buffers for the whole thing on st 0 root only       */
  int took;               /* result of the implicit Take which initiates a stage */
  int outbuffspaceleft;   /* during output, how much space is left in the out buffer */
  int outbufflensspaceleft;   /* during output, how much space is left in the outlens buffer */
  char *sndptr, *rcvptr;
  int totalsendlen;       /* for SPLOCAL set up, total lengths of a P's contributions */

  MPI_Group stagerootonlygroup, tempgroup;

  pthread_t sendthread;
  pthread_t recvthread;
  void *threadval;      

  eSkel_atom_t *tempitemin;
  eSkel_atom_t *tempitemout;

  int w, x; char *wptr;   /* debugging only */

  Top++;
  /* Some of this copying is probably redundant */
  FrStack[Top] = (skel_t *) malloc (sizeof(skel_t));
  FrStack[Top]->skel = Pskel;
  FrStack[Top]->context.p.ns    = ns;

  FrStack[Top]->context.p.parentcomm = comm;
  MPI_Comm_size(comm, &FrStack[Top]->context.p.parentcommsize);
  MPI_Comm_rank(comm, &FrStack[Top]->context.p.parentcommrank);

  FrStack[Top]->context.p.spr        = spr;  
  FrStack[Top]->context.p.ty         = ty;   
  FrStack[Top]->context.p.inmul      = inmul;
  FrStack[Top]->context.p.mystage    = col;
  FrStack[Top]->context.p.itemsz     = (int *) malloc (sizeof(int)*ns+1);
  for (i=0; i<ns+1; i++) {
    MPI_Type_size(ty[i], &FrStack[Top]->context.p.itemsz[i]);
  }
  MPI_Comm_split(comm, col, 0, &FrStack[Top]->mycomm);
  MPI_Comm_rank(FrStack[Top]->mycomm, &FrStack[Top]->myrank);
  MPI_Comm_size(FrStack[Top]->mycomm, &FrStack[Top]->mycommsize);

  /* work out who's who in interesting stages */
  FrStack[Top]->context.p.allcols 
       = (int *) malloc (sizeof(int)*FrStack[Top]->context.p.parentcommsize);
  MPI_Allgather((void *) &FrStack[Top]->context.p.mystage, 1, MPI_INT,
                (void *) FrStack[Top]->context.p.allcols, 1, MPI_INT, comm);
  FrStack[Top]->context.p.nextstage 
       = (int *) malloc (sizeof(int)*FrStack[Top]->context.p.parentcommsize);
  FrStack[Top]->context.p.prevstage 
       = (int *) malloc (sizeof(int)*FrStack[Top]->context.p.parentcommsize);
  FrStack[Top]->context.p.firststage 
       = (int *) malloc (sizeof(int)*FrStack[Top]->context.p.parentcommsize);
  FrStack[Top]->context.p.laststage 
       = (int *) malloc (sizeof(int)*FrStack[Top]->context.p.parentcommsize);
  FrStack[Top]->context.p.thisstage 
       = (int *) malloc (sizeof(int)*FrStack[Top]->context.p.parentcommsize);

  FrStack[Top]->context.p.firststageroot = (int *) malloc (sizeof(int));
  FrStack[Top]->context.p.firststageroot[0] = -1;
  
  FrStack[Top]->context.p.laststageroot = (int *) malloc (sizeof(int));
  FrStack[Top]->context.p.laststageroot[0] = -1;

  j=0; k=0; l=0; m=0; n=0;
  for (i=0; i<FrStack[Top]->context.p.parentcommsize; i++) {
    if (FrStack[Top]->context.p.allcols[i] == col+1) 
      FrStack[Top]->context.p.nextstage[j++]=i;
    if (FrStack[Top]->context.p.allcols[i] == 0) 
      FrStack[Top]->context.p.firststage[k++]=i;
    if (FrStack[Top]->context.p.allcols[i] == ns-1) 
      FrStack[Top]->context.p.laststage[l++]=i;
    if (FrStack[Top]->context.p.allcols[i] == col) 
      FrStack[Top]->context.p.thisstage[m++]=i;
    if (FrStack[Top]->context.p.allcols[i] == col-1) 
      FrStack[Top]->context.p.prevstage[n++]=i;
    if (FrStack[Top]->context.p.allcols[i] == 0 && FrStack[Top]->context.p.firststageroot[0]==-1)
      FrStack[Top]->context.p.firststageroot[0]=i;
    if (FrStack[Top]->context.p.allcols[i] == ns-1 && FrStack[Top]->context.p.laststageroot[0]==-1)
      FrStack[Top]->context.p.laststageroot[0]=i;
  }
  FrStack[Top]->context.p.nextstagesize  = j;
  FrStack[Top]->context.p.firststagesize = k;
  FrStack[Top]->context.p.laststagesize  = l;
  FrStack[Top]->context.p.thisstagesize  = m;
  FrStack[Top]->context.p.prevstagesize  = n;
  
  MPI_Comm_dup(FrStack[Top]->context.p.parentcomm, &(FrStack[Top]->context.p.allcomm1));
  MPI_Comm_dup(FrStack[Top]->context.p.parentcomm, &(FrStack[Top]->context.p.allcomm2));
  MPI_Comm_dup(FrStack[Top]->mycomm, &(FrStack[Top]->context.p.thisstagecomm));

  MPI_Comm_group(FrStack[Top]->mycomm, &tempgroup);
  MPI_Group_incl(tempgroup,1,&zero, &stagerootonlygroup);
  MPI_Comm_create(FrStack[Top]->mycomm, stagerootonlygroup, &(FrStack[Top]->context.p.thisstagerootcomm));

  /* Now gather all the input to a buffer in stage 0. A more sophisticated  */
  /* implementation will leave this where it is, and a correspondingly more */
  /* complex "Take" operation will know how to deal with it                 */

  FrStack[Top]->context.p.allranks = (int *) malloc (sizeof(int)* FrStack[Top]->context.p.parentcommsize);
  for (i=0; i<FrStack[Top]->context.p.parentcommsize; i++) {
    FrStack[Top]->context.p.allranks[i] = i;
  }
  if (spr[0] == SPGLOBAL) {
    sndptr = (char *) in;
    if (col==0) {
      /* Set up the durable stage 0 input buffers */
      FrStack[Top]->context.p.blens = (int *)  malloc (sizeof(int)*inmul);
      FrStack[Top]->context.p.buff  = (char **)malloc (sizeof(char *)*inmul);
    }
    /* Now the "senders" loop calling group2group */
    /* spark receiver threads on stage O processes */
    if (col == 0) {
      pthread_create(&recvthread, NULL, (void* (*)(void *))&piperecvthread, &inmul);
    }
    for (i=0; i<inmul; i++) {
      group2group (FrStack[Top]->context.p.allcomm1, FrStack[Top]->context.p.parentcommsize,
                   FrStack[Top]->context.p.parentcommrank, SENDER,  FrStack[Top]->context.p.parentcommrank,
                   FrStack[Top]->context.p.allcomm2,  FrStack[Top]->context.p.parentcommsize, 
                   FrStack[Top]->context.p.allranks,
                   FrStack[Top]->context.p.thisstagecomm, FrStack[Top]->mycommsize,  
                   FrStack[Top]->context.p.firststage,
                   FrStack[Top]->context.p.ty[0], FrStack[Top]->context.p.itemsz[0],
                   sndptr, inlen[i], NULL, NULL);
      sndptr = sndptr + (FrStack[Top]->context.p.itemsz[0] * inlen[i]);
    }
    FrStack[Top]->context.p.bnext = 0; /* set up "pointer" for Take */
    if (col == 0) {
      pthread_join(recvthread, &threadval);
    }
  } else { /* gathering local spread data */
    /* spark receiver thread on stage O root */
    if (col == 0 && FrStack[Top]->myrank == 0) {
      pthread_create(&recvthread, NULL, (void* (*)(void *))&piperootrecvthread, NULL);
    }
    /* send the lengths with a g2g */
    group2group (FrStack[Top]->context.p.allcomm1, FrStack[Top]->context.p.parentcommsize,
                 FrStack[Top]->context.p.parentcommrank, SENDER,  FrStack[Top]->context.p.parentcommrank,
                 FrStack[Top]->context.p.allcomm2,  FrStack[Top]->context.p.parentcommsize, 
                 FrStack[Top]->context.p.allranks,
                 FrStack[Top]->context.p.thisstagerootcomm, 1,  
                 FrStack[Top]->context.p.firststageroot,
                 MPI_INT, mpiintsize, (char *)inlen, inmul, NULL, NULL);
    
    /* count and send the data with g2g */
    for (totalsendlen=0, i=0; i<inmul; i++) {
      totalsendlen += inlen[i];
    }
    group2group (FrStack[Top]->context.p.allcomm1, FrStack[Top]->context.p.parentcommsize,
                 FrStack[Top]->context.p.parentcommrank, SENDER,  FrStack[Top]->context.p.parentcommrank,
                 FrStack[Top]->context.p.allcomm2,  FrStack[Top]->context.p.parentcommsize, 
                 FrStack[Top]->context.p.allranks,
                 FrStack[Top]->context.p.thisstagerootcomm, 1,  
                 FrStack[Top]->context.p.firststageroot,
                 FrStack[Top]->context.p.ty[0], FrStack[Top]->context.p.itemsz[0], 
                 in, totalsendlen, NULL, NULL);
    if (col == 0 && FrStack[Top]->myrank == 0) {
      pthread_join(recvthread, &threadval);
    }
  }

  /* Now set up the temporary output buffers in the final stage */
  /* Need to handle unknown number of outputs                   */
  /* Find stage ns-1 root as comm id -> outbuffroot*/
  if (col == ns-1) {
    temp = FrStack[Top]->context.p.parentcommrank;
  } else {
    temp = FrStack[Top]->context.p.parentcommsize;
  }
  MPI_Allreduce(&temp, &outbuffroot, 1, MPI_INT, MPI_MIN, comm);
  if (col == ns-1) {
    if (spr[ns] == SPGLOBAL) {
        FrStack[Top]->context.p.poutblens = (int *)  malloc (sizeof(int)*MAXOUTMUL);
        FrStack[Top]->context.p.poutbuff  = (char **)malloc (sizeof(char *)*MAXOUTMUL);
        FrStack[Top]->context.p.poutbnext = 0;
    } else if (spr[ns] == SPLOCAL) {
      if (FrStack[Top]->context.p.parentcommrank == outbuffroot) {
        FrStack[Top]->context.p.poutblens = (int *)  malloc (sizeof(int)*MAXOUTMUL);
        FrStack[Top]->context.p.poutbuff  = (char **)malloc (sizeof(char *)*MAXOUTMUL);
        FrStack[Top]->context.p.poutbnext = 0;
      }
    }
  }

  /* Run the pipeline */

  FrStack[Top]->context.p.shutdown = FALSE;
  tempitemin = (eSkel_atom_t *) malloc (sizeof (eSkel_atom_t));
  while ( tempitemin->data = Take (&(tempitemin->len)))  {
    tempitemout = stages[col](tempitemin);
    if (tempitemout) Give (tempitemout->data, tempitemout->len);
    else Give (NULL, 0);
  }
  if (col < ns-1) Give(NULL, SHUTDOWNLEN);


  /* Finally distribute output from buffer at final stage */
  if (spr[ns] == SPGLOBAL) {
    outbuffspaceleft     = outlen[0];
    outbufflensspaceleft = outlen[1]; 
    /* broadcast actual multiplicity of output */
    MPI_Bcast(&FrStack[Top]->context.p.poutbnext, 1, MPI_INT, FrStack[Top]->context.p.laststage[0], comm);
    FrStack[Top]->context.p.blens = (int *) malloc (sizeof(int)* FrStack[Top]->context.p.poutbnext);
    rcvptr = (char *) out;
    /* Now the "receivers" loop calling group2group */
    /* spark sender threads on stage ns-1 processes */
    if (col == ns-1) {
      pthread_create(&sendthread, NULL, (void* (*)(void *))&pipesendthread, &(FrStack[Top]->context.p.poutbnext));
    }
    for (i=0; i<FrStack[Top]->context.p.poutbnext; i++) {
      temp = outbuffspaceleft;
      group2group (FrStack[Top]->context.p.allcomm1, FrStack[Top]->context.p.parentcommsize,
                   FrStack[Top]->context.p.parentcommrank, RECEIVER,  FrStack[Top]->context.p.parentcommrank,
                   FrStack[Top]->context.p.thisstagecomm,  FrStack[Top]->mycommsize,  
                   FrStack[Top]->context.p.laststage,
                   FrStack[Top]->context.p.allcomm2, FrStack[Top]->context.p.parentcommsize, 
                   FrStack[Top]->context.p.allranks,
                   FrStack[Top]->context.p.ty[ns], FrStack[Top]->context.p.itemsz[ns],
                   NULL, 0, &rcvptr, &temp);
      outbuffspaceleft -= temp;
      if (i<outbufflensspaceleft) {
        outlen[i] = temp;
      } else {
       fprintf(stderr, "No more space for pipeline outbufflens\n");
       exit(0);
      }
      rcvptr += (temp*FrStack[Top]->context.p.itemsz[ns]);
    }
    if (col == ns-1) {
      pthread_join(sendthread, &threadval);
    }
    *outmul = FrStack[Top]->context.p.poutbnext;
  } else {    /* distribute local spread results */
    /* It would be "neat" to do this with g2g as well, but it just doesn't fit because of the */
    /* possibility of varying item length in pipelinev. If the data length is uniform, eg as  */
    /* in pipeline, then it would fit.                                                        */

    Local_Mult_Scatter (comm, FrStack[Top]->context.p.parentcommsize, FrStack[Top]->context.p.parentcommrank,
                        outbuffroot, ty[ns], FrStack[Top]->context.p.itemsz[ns],
                        (void **) FrStack[Top]->context.p.poutbuff , FrStack[Top]->context.p.poutbnext, 
                        FrStack[Top]->context.p.poutblens, (char *) out, outmul, outlen);
  }
  Top--; /* and give back the space to be tidy */
  return;
}


